/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.gogrid.compute;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.MachineImageFormat;
import org.dasein.cloud.compute.MachineImageState;
import org.dasein.cloud.compute.MachineImageSupport;
import org.dasein.cloud.compute.MachineImageType;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.jclouds.gogrid.GoGrid;
import org.jclouds.gogrid.GoGridAsyncClient;
import org.jclouds.gogrid.GoGridClient;
import org.jclouds.gogrid.GoGridResponseException;
import org.jclouds.gogrid.domain.ServerImage;
import org.jclouds.gogrid.options.GetImageListOptions;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.ResourceNotFoundException;
import org.jclouds.rest.RestContext;

public class GoGridImage implements MachineImageSupport {
    static private final Logger logger = Logger.getLogger(GoGridImage.class);
    
    private GoGrid provider;
    
    GoGridImage(GoGrid gogrid) {
        provider = gogrid;
    }

    @Override
    public void remove(String imageId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not allow you to delete images.");
    }

    @Override
    public Collection<MachineImage> searchMachineImages(String keyword, Platform platform, Architecture architecture) throws InternalException, CloudException {
        try {
            ArrayList<MachineImage> list = new ArrayList<MachineImage>();
            Iterable<MachineImage> raw;
            
            try {
                raw = listMachineImages();
            }
            catch( RuntimeException e ) {
                logger.warn("Error listing private images for search: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
            for( MachineImage image :  raw ) {
                if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                    continue;
                }
                if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                    Platform mine = image.getPlatform();
                    
                    if( platform.isWindows() && !mine.isWindows() ) {
                        continue;
                    }
                    if( platform.isUnix() && !mine.isUnix() ) {
                        continue;
                    }
                    if( platform.isBsd() && !mine.isBsd() ) {
                        continue;
                    }
                    if( platform.isLinux() && !mine.isLinux() ) {
                        continue;
                    }
                    if( platform.equals(Platform.UNIX) ) {
                        if( !mine.isUnix() ) {
                            continue;
                        }
                    }
                    else if( !platform.equals(mine) ) {
                        continue;
                    }
                }
                if( keyword != null && !keyword.equals("") ) {
                    keyword = keyword.toLowerCase();
                    if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                        if( !image.getName().toLowerCase().contains(keyword) ) {
                            if( !image.getDescription().toLowerCase().contains(keyword) ) {
                                continue;
                            }
                        }
                    }
                }
                list.add(image);
            }
            try {
                raw = listMachineImagesOwnedBy(null);
            }
            catch( RuntimeException e ) {
                logger.warn("Error listing public images for search: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);            
            }
            for( MachineImage image : raw ) {
                System.out.println("Checking public image: " + image);
                System.out.println("Terms: " + keyword + "/" + architecture + "/" + platform);
                System.out.println("image owner=" + image.getProviderOwnerId());
                if( provider.getContext().getAccountNumber().equals(image.getProviderOwnerId()) ) {
                    continue;
                }
                if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                    continue;
                }
                if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                    Platform mine = image.getPlatform();
                    
                    if( platform.isWindows() && !mine.isWindows() ) {
                        continue;
                    }
                    if( platform.isUnix() && !mine.isUnix() ) {
                        continue;
                    }
                    if( platform.isBsd() && !mine.isBsd() ) {
                        continue;
                    }
                    if( platform.isLinux() && !mine.isLinux() ) {
                        continue;
                    }
                    if( platform.equals(Platform.UNIX) ) {
                        if( !mine.isUnix() ) {
                            continue;
                        }
                    }
                    else if( !platform.equals(mine) ) {
                        continue;
                    }
                }
                if( keyword != null && !keyword.equals("") ) {
                    keyword = keyword.toLowerCase();
                    if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                        if( !image.getName().toLowerCase().contains(keyword) ) {
                            if( !image.getDescription().toLowerCase().contains(keyword) ) {
                                continue;
                            }
                        }
                    }
                }
                list.add(image);
            }
            return list;
        }
        catch( RuntimeException e ) {
            logger.error("Runtime error during image search: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
        catch( Error e ) {
            logger.error("Error during image search: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        } 
    }

    @Override
    public MachineImage getMachineImage(String imageId) throws InternalException, CloudException {
        return toImage(getGoGridImage(imageId));
    }

    private ServerImage getGoGridImage(String imageId) throws CloudException, InternalException {
        try {
            RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
            
            try {
                long id;
                
                try {
                    id = Long.parseLong(imageId);
                }
                catch( NumberFormatException e ) {
                    return null;
                }
                try {
                    for( ServerImage image : ctx.getApi().getImageServices().getImagesById(id) ) {
                        return image;
                    }
                    return null;
                }
                catch( ResourceNotFoundException e ) {
                    return null;
                }
                catch( GoGridResponseException e ) {
                    logger.error("Failed to lookup image " + imageId + ": " + e.getMessage());
                    if( logger.isDebugEnabled() ) {
                        e.printStackTrace();
                    }
                    throw new CloudException("Failed to lookup image " + imageId + ": " + e.getMessage());            
                }
            }
            finally {
                ctx.close();
            }
        }
        catch( RuntimeException e ) {
            logger.error("Failed to lookup image " + imageId + ": " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException("Failed to lookup image " + imageId + ": " + e.getMessage());            
        }
        catch( Error e ) {
            logger.error("Failed to lookup image " + imageId + ": " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException("Failed to lookup image " + imageId + ": " + e.getMessage());            
        }
    }

    @Override
    public String getProviderTermForImage(Locale locale) {
        return "image";
    }

    @Override
    public Collection<String> listShares(String imageId) throws CloudException, InternalException {
        return new ArrayList<String>();
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachine(String serverId, String name, String description) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    @Override
    public Collection<MachineImage> listMachineImages() throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            ArrayList<MachineImage> images = new ArrayList<MachineImage>();
            GetImageListOptions options = new GetImageListOptions();
            
            options = options.inDatacenter(provider.getContext().getRegionId()).onlyPrivate();
            for( ServerImage image : ctx.getApi().getImageServices().getImageList(options) ) {
                MachineImage img = toImage(image);
                
                if( img != null ) {
                    images.add(img);
                }
            }
            return images;
        }
        catch( RuntimeException e ) {
            logger.error("Failed to list images: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to list images: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }
    
    private MachineImage toImage(ServerImage goGridImage) throws InternalException, CloudException {
        if( goGridImage == null ) {
            return null;
        }
        MachineImage image = new MachineImage();
        
        image.setCurrentState(goGridImage.isActive() ? MachineImageState.ACTIVE : MachineImageState.PENDING);
        image.setArchitecture(provider.toArchitecture(goGridImage.getOs(), goGridImage.getDescription()));

        image.setDescription(goGridImage.getDescription());
        image.setName(goGridImage.getFriendlyName());
        image.setProviderOwnerId(String.valueOf(goGridImage.getOwner().getId()));
        image.setPlatform(provider.toPlatform(goGridImage.getOs()));
        long ownerId = goGridImage.getOwner().getId();
        
        image.setProviderMachineImageId(ownerId == -1L ? "-1" : goGridImage.getOwner().getName());
        image.setSoftware("");
        image.setType(MachineImageType.STORAGE);
        image.setProviderRegionId(provider.getContext().getRegionId());
        return image;
    }

    @Override
    public void downloadImage(String machineImageId, OutputStream toOutput) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean hasPublicLibrary() {
        return true;
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachineToStorage(String vmId, String name, String description, String directory) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public String installImageFromUpload(MachineImageFormat format, InputStream imageStream) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean isImageSharedWithPublic(String machineImageId) throws CloudException, InternalException {
        return false;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            GetImageListOptions options = new GetImageListOptions();
            
            options = options.inDatacenter(provider.getContext().getRegionId()).onlyPublic();
            ctx.getApi().getImageServices().getImageList(options);
            return true;
        }
        catch( AuthorizationException e ) {
            return false;
        }
        catch( RuntimeException e ) {
            logger.error("Failed to list images: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to list images: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public Iterable<MachineImage> listMachineImagesOwnedBy(String accountId) throws CloudException, InternalException {
        if( accountId == null ) {
            RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
            
            try {
                ArrayList<MachineImage> images = new ArrayList<MachineImage>();
                GetImageListOptions options = new GetImageListOptions();
                
                options = options.inDatacenter(provider.getContext().getRegionId()).onlyPublic();
                for( ServerImage image : ctx.getApi().getImageServices().getImageList(options) ) {
                    MachineImage img = toImage(image);
                    
                    if( img != null ) {
                        images.add(img);
                    }
                }
                return images;
            }
            catch( RuntimeException e ) {
                logger.error("Failed to list images: " + e.getMessage());
                e.printStackTrace();
                throw new CloudException("Failed to list images: " + e.getMessage());
            }
            finally {
                ctx.close();
            }
        }
        else if( accountId.equals(provider.getContext().getAccountNumber()) ) {
            return listMachineImages();
        }
        else {
            return Collections.emptyList();
        }
    }

    @Override
    public Iterable<MachineImageFormat> listSupportedFormats() throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public String registerMachineImage(String atStorageLocation) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No supported");
    }

    @Override
    public void shareMachineImage(String machineImageId, String withAccountId, boolean allow) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No supported");
    }

    @Override
    public boolean supportsCustomImages() {
        return false;
    }

    @Override
    public boolean supportsImageSharing() {
        return false;
    }

    @Override
    public boolean supportsImageSharingWithPublic() {
        return false;
    }

    @Override
    public String transfer(CloudProvider fromCloud, String machineImageId) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No supported");
    }
}
