/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.gogrid.compute;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;
import java.util.Set;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.Tag;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.compute.VirtualMachineProduct;
import org.dasein.cloud.compute.VirtualMachineSupport;
import org.dasein.cloud.compute.VmState;
import org.dasein.cloud.compute.VmStatistics;
import org.dasein.cloud.jclouds.gogrid.GoGrid;
import org.dasein.cloud.network.IpAddress;
import org.dasein.util.CalendarWrapper;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;
import org.jclouds.domain.Credentials;
import org.jclouds.gogrid.GoGridAsyncClient;
import org.jclouds.gogrid.GoGridClient;
import org.jclouds.gogrid.GoGridResponseException;
import org.jclouds.gogrid.domain.Ip;
import org.jclouds.gogrid.domain.Option;
import org.jclouds.gogrid.domain.PowerCommand;
import org.jclouds.gogrid.domain.Server;
import org.jclouds.gogrid.options.AddServerOptions;
import org.jclouds.gogrid.options.GetServerListOptions;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.RestContext;

public class GoGridServer implements VirtualMachineSupport {
    static private final Logger logger = Logger.getLogger(GoGridServer.class);
    
    private final GoGrid provider;
    
    GoGridServer(GoGrid gogrid) {
        provider = gogrid;
    }
    
    @Override
    public void boot(String serverId) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {   
            ctx.getApi().getServerServices().power(serverId, PowerCommand.START);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to power on server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to power on server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public VirtualMachine clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support server cloning.");
    }

    @Override
    public String getConsoleOutput(String serverId) throws InternalException, CloudException {
        return "";
    }

    @Override
    public Collection<String> listFirewalls(String serverId) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    @Override
    public String getProviderTermForServer(Locale locale) {
        return "server";
    }

    @Override
    public VirtualMachine getVirtualMachine(String serverId) throws InternalException, CloudException {
        return toServer(getGoGridServer(serverId));
    }

    private Server getGoGridServer(String serverId) throws CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        long id;
        
        try {
            id = Long.parseLong(serverId);
        }
        catch( NumberFormatException e ) {
            return null;
        }
        try {
            for( org.jclouds.gogrid.domain.Server server : ctx.getApi().getServerServices().getServersById(id) ) {
                return server;
            }        
        }
        catch( RuntimeException e ) {
            String msg = e.getMessage();
            
            if( msg.contains("No object found") ) {
                return null;
            }
            logger.error("Failed to find server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to find server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
        return null;
    }

    @Override
    public VirtualMachineProduct getProduct(String productId) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        long id;
        
        try {
            id = Long.parseLong(productId);
        }
        catch( NumberFormatException e ) {
            return null;
        }
        try {
            Set<Option> sizes = ctx.getApi().getServerServices().getRamSizes();
            
            for( Option size : sizes ) {
                if( size.getId() == id ) {
                    VirtualMachineProduct product = new VirtualMachineProduct();
                    
                    product.setProductId(String.valueOf(size.getId()));
                    product.setName(size.getName());
                    product.setDescription(size.getDescription());
                    product.setCpuCount(1);
                    product.setDiskSizeInGb(50);
                    String s = size.getName().substring(0, size.getName().length()-2);
                    product.setRamInMb(Integer.parseInt(s));
                    return product;
                }
            }
            return null;
        }
        catch( Throwable t ) {
            throw new CloudException(t);
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public VmStatistics getVMStatistics(String serverId, long begin, long end) throws InternalException, CloudException {
        return new VmStatistics();
    }

    @Override
    public Collection<VmStatistics> getVMStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return Collections.emptyList();
    }
    
    @Override
    public Iterable<VirtualMachineProduct> listProducts(Architecture architecture) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            Set<Option> sizes = ctx.getApi().getServerServices().getRamSizes();
            ArrayList<VirtualMachineProduct> list = new ArrayList<VirtualMachineProduct>();
            
            for( Option size : sizes ) {
                VirtualMachineProduct product = new VirtualMachineProduct();
                
                product.setProductId(String.valueOf(size.getId()));
                product.setName(size.getName());
                product.setDescription(size.getDescription());
                product.setCpuCount(1);
                product.setDiskSizeInGb(1);
                product.setRamInMb(256);
                list.add(product);
            }
            return list;
        }
        catch( Throwable t ) {
            throw new CloudException(t);
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public VirtualMachine launch(String imageId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String... firewalls) throws InternalException, CloudException {
        return launch(imageId, size, dataCenterId, serverName, description, withKey, inVlanId, withMonitoring, forImaging, firewalls, new Tag[0]);
    }
    
    public VirtualMachine launch(String imageId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String[] firewalls, Tag ... tags) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            ArrayList<Long> ids = new ArrayList<Long>();
            boolean waiting = true;
            IpAddress address = null;
            
            serverName = validateServerName(ctx, serverName);
            // GoGrid doesn't return a server ID immediately on addServer, so we have to figure it out
            while( waiting ) {
                waiting = false;
                try {
                    for( Server server : ctx.getApi().getServerServices().getServersByName(serverName) ) {
                        if( server.getId() == 0L ) {
                            waiting = true;
                            ids.clear();
                            break;
                        }
                        ids.add(server.getId());
                    }
                }
                catch( GoGridResponseException ignore ) {
                    // ignore this
                }
                if( waiting ) {
                    try { Thread.sleep(10000L); }
                    catch( InterruptedException e ) { }
                }
            }
            for( IpAddress ip : provider.getNetworkServices().getIpAddressSupport().listPublicIpPool(true) ) {
                address = ip;
            }
            if( address == null ) {
                throw new CloudException("No more addresses are left in this account.");
            }
            org.jclouds.gogrid.domain.Server server;
            
            try {   
                AddServerOptions opts = new AddServerOptions();
                
                if( forImaging ) {
                    opts = opts.asSandboxType().withDescription(description);
                }
                else {
                    opts = opts.withDescription(description);
                }
                server = ctx.getApi().getServerServices().addServer(serverName, imageId, size.getProductId(), address.getProviderIpAddressId(), opts);
            }
            catch( GoGridResponseException e ) {
                String msg = e.getMessage();
                
                if( msg != null && msg.contains("IP already in use") ) {
                    return launch(imageId, size, dataCenterId, serverName, description, withKey, inVlanId, withMonitoring, forImaging, firewalls);
                }
                throw new CloudException(e.getMessage());
            }
            if( server == null ) {
                throw new CloudException("Failed to create server " + serverName);
            }
            server = null;
            while( server == null ) {
                try {
                    for( org.jclouds.gogrid.domain.Server s : ctx.getApi().getServerServices().getServersByName(serverName) ) {
                        long id = s.getId();
                        
                        if( id != 0L && !ids.contains(id) ) {
                            server = s;
                            break;
                        }
                    }
                }
                catch( GoGridResponseException ignore ) {
                    // waiting
                }
                if( server == null ) {
                    try { Thread.sleep(10000L); }
                    catch( InterruptedException e ) { }
                }
            }
            return toServer(server);
        }
        catch( RuntimeException e ) {
            throw new CloudException("Failed to define server: " + e.getMessage());
        }
        catch( Error e ) {
            throw new InternalException("Failed to define server: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public Collection<VirtualMachine> listVirtualMachines() throws InternalException, CloudException {
        provider.hold();
        PopulatorThread<VirtualMachine> populator = new PopulatorThread<VirtualMachine>(new JiteratorPopulator<VirtualMachine>() {
            @Override
            public void populate(Jiterator<VirtualMachine> iterator) throws Exception {
                try {
                    RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
                    
                    try {
                        try {
                            GetServerListOptions options = new GetServerListOptions();
                            
                            options.inDatacenter(provider.getContext().getRegionId());
                            for( Server server : ctx.getApi().getServerServices().getServerList(options) ) {
                                if( server != null && server.getId() > 0 ) {
                                    VirtualMachine vm = toServer(server);

                                    if( vm.getCurrentState().equals(VmState.RUNNING) ) {
                                        Credentials credentials = ctx.getApi().getServerServices().getServerCredentialsList().get(server.getName());
                                        if( credentials != null ) {
                                            vm.setRootUser(credentials.identity);
                                            vm.setRootPassword(credentials.credential);
                                        }
                                    }
                                    iterator.push(vm);
                                }
                            }
                        }
                        catch( RuntimeException e ) {
                            logger.error("Failed to list servers: " + e.getMessage());
                            e.printStackTrace();
                            throw new CloudException("Failed to list servers: " + e.getMessage());
                        }
                    }
                    finally {
                        ctx.close();
                    }
                }
                finally {
                    try { provider.release(); }
                    catch( Throwable t ) { t.printStackTrace(); }
                }
            }            
        });

        populator.populate();
        return populator.getResult();
    }

    @Override
    public void enableAnalytics(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().getServerServices().power(serverId, PowerCommand.STOP);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to pause server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to pause server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().getServerServices().power(serverId, PowerCommand.RESTART);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to reboot server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to reboot server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void terminate(String serverId) throws InternalException, CloudException {
        try {
            pause(serverId);
            long timeout = System.currentTimeMillis() + (CalendarWrapper.MINUTE*15L);
            VirtualMachine vm = getVirtualMachine(serverId);
            
            while( !vm.getCurrentState().equals(VmState.PAUSED) ) {
                if( System.currentTimeMillis() >= timeout ) {
                    break;
                }
                try { Thread.sleep(15000L); }
                catch( InterruptedException e ) { }
                try { vm = getVirtualMachine(serverId); }
                catch( Throwable ignore ) { }
            }
            RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
    
            try {
                ctx.getApi().getServerServices().deleteById(Long.parseLong(serverId));        
            }
            catch( GoGridResponseException e ) {
                logger.error("Failed to delete " + serverId + ": " + e.getMessage());
                e.printStackTrace();
                throw new CloudException("Failed to delete " + serverId + ": " + e.getMessage());
            }
            finally {
                ctx.close();
            }
        }
        catch( RuntimeException e ) {
            logger.error("Runtime error trying to terminate " + serverId + ": " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }    
        catch( Error e ) {
            logger.error("Error trying to terminate " + serverId + ": " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
    }

    @Override
    public void disableAnalytics(String serverId) throws InternalException, CloudException {
        // NO-OP
    }
    
    private VirtualMachineProduct toSize(Option ram) throws InternalException, CloudException {
        return getProduct(String.valueOf(ram.getId()));
    }
    

    private VmState toState(org.jclouds.gogrid.domain.ServerState serverState) {
       //   ON, STARTING, OFF, STOPPING, RESTARTING, SAVING, RESTORING, UPDATING, UNRECOGNIZED;
        if(serverState ==  null|| serverState ==  org.jclouds.gogrid.domain.ServerState.UNRECOGNIZED || serverState == org.jclouds.gogrid.domain.ServerState.STARTING ) { 
            return VmState.PENDING;
        }
        switch( serverState ) {
            case ON: return VmState.RUNNING; 
            case OFF: return VmState.PAUSED;
            case STOPPING: return VmState.STOPPING;
            case RESTARTING: return VmState.REBOOTING;
            case SAVING: return VmState.REBOOTING;
            case RESTORING: return VmState.PENDING;
        }
        return VmState.RUNNING;
    }
    
    private VirtualMachine toServer(org.jclouds.gogrid.domain.Server fromGoGrid) throws InternalException, CloudException {
        if( fromGoGrid != null && fromGoGrid.getId() > 0 ) {
            org.jclouds.gogrid.domain.ServerImage image = fromGoGrid.getImage();
            VirtualMachine server = new VirtualMachine();
            
            server.setName(fromGoGrid.getName());
            server.setProviderVirtualMachineId(String.valueOf(fromGoGrid.getId()));
            server.setProviderMachineImageId(String.valueOf(image.getId()));
            server.setArchitecture(provider.toArchitecture(image.getOs(), fromGoGrid.getDescription()));
            Ip ip = fromGoGrid.getIp();
            
            if( ip != null ) {
                if( ip.isPublic() ) {
                    server.setPublicIpAddresses(new String[] { ip.getIp() });
                    server.setProviderAssignedIpAddressId(String.valueOf(ip.getId()) );
                }
                else {
                    server.setPrivateIpAddresses(new String[] { ip.getIp() });
                }
            }
            server.setCreationTimestamp(0L);
            server.setCurrentState(toState(fromGoGrid.getState()));
            server.setProviderDataCenterId(fromGoGrid.getDatacenter().getId() + "a");
            server.setDescription(fromGoGrid.getDescription());
            server.setPersistent(true);
            server.setPlatform(provider.toPlatform(fromGoGrid.getOs()));
            server.setProviderRegionId(String.valueOf(fromGoGrid.getDatacenter().getId()));
            VirtualMachineProduct prd = toSize(fromGoGrid.getRam());
            
            if( prd == null ) {
                prd = new VirtualMachineProduct();
               
                long id = fromGoGrid.getRam().getId();
                String name = fromGoGrid.getRam().getName();
                String desc = fromGoGrid.getRam().getDescription();
                
                prd.setProductId(String.valueOf(id));
                prd.setName(name == null ? String.valueOf(id) : name);
                prd.setDescription(desc == null ? (name == null ? String.valueOf(id) : name) : desc);
                prd.setCpuCount(1);
                prd.setDiskSizeInGb(50);
                try {
                    String s = prd.getName().substring(0, prd.getName().length()-2);
                    prd.setRamInMb(Integer.parseInt(s));
                }
                catch( Throwable t ) {
                    prd.setRamInMb(0);
                }
            }
            server.setProduct(prd);
            server.setLastBootTimestamp(0L);
            server.setLastPauseTimestamp(0L);
            server.setTerminationTimestamp(-1L);
            server.setProviderOwnerId(provider.getContext().getAccountNumber());
            return server;
        }
        return null;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            try {
                GetServerListOptions options = new GetServerListOptions();
                
                options.inDatacenter(provider.getContext().getRegionId());
                ctx.getApi().getServerServices().getServerList(options);
                return true;
            }
            catch( AuthorizationException e ) {
                return false;
            }
            catch( RuntimeException e ) {
                logger.error("Failed to determine subscription status: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException("Failed to determine subscription status: " + e.getMessage());
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public boolean supportsAnalytics() throws CloudException, InternalException {
        return false;
    }
    
    private String validateServerName(RestContext<GoGridClient,GoGridAsyncClient> ctx, String serverName) {
        if( serverName == null ) {
            serverName = String.valueOf(System.currentTimeMillis());
        }
        if( serverName.length() > 20 ) {
            serverName = serverName.substring(0,20);
        }
        try {
            String name = serverName;
            int postfix = 1;
            boolean found;
            
            do {
                found = false;
                for( Server server : ctx.getApi().getServerServices().getServersByName(name) ) {
                    if( server != null ) {
                        name = serverName + (postfix++);
                        found = true;
                        break;
                    }
                }
            } while( found );
            return name;
        }
        catch( GoGridResponseException ignore ) {
            // ignore this
        }      
        return serverName;
    }
}
